use arrayvec::ArrayString;
use core::convert::Infallible;
use core::fmt::Write;
use cortex_m::asm;
use cortex_m_rt::{entry, exception};
use embedded_graphics::{
    drawable::Drawable, fonts::Text, image::Image, pixelcolor::BinaryColor, pixelcolor::Rgb565,
    prelude::*, primitives::rectangle::Rectangle, style::PrimitiveStyleBuilder,
    style::TextStyleBuilder,
};
use embedded_hal::blocking::spi;
use embedded_hal::digital::v2::{InputPin, OutputPin};
use embedded_hal::Qei;
use profont::{ProFont12Point, ProFont9Point};
use rtt_target::{rprintln, rtt_init_print};
use st7735_lcd::Orientation;
use stm32f1xx_hal::{
    delay::Delay,
    gpio::{gpioa, gpiob, gpioc, Alternate, Floating, Input, Output, PushPull},
    pac,
    prelude::*,
    qei, rcc,
    spi::{Mode, Phase, Polarity, Spi, Spi1NoRemap},
    stm32,
    timer::{Tim3PartialRemap, Timer},
};
use tinybmp::Bmp;

pub struct App<BLed, SPI, DispCS, DispDC, DispRST, MaxCS, QEI>
where
    BLed: OutputPin,
    SPI: spi::Transfer<u8> + spi::Write<u8>,
    DispCS: OutputPin,
    DispDC: OutputPin,
    DispRST: OutputPin,
    MaxCS: OutputPin,
    QEI: Qei,
{
    delay: Delay,
    board_led: BLed,
    spi: SPI,
    disp_cs: DispCS,
    disp_dc: DispDC,
    disp_rst: DispRST,
    max_cs: MaxCS,
    qei: QEI,
}

pub fn setup(
    cp: cortex_m::peripheral::Peripherals,
    dp: stm32::Peripherals,
) -> App<
    gpioc::PC13<Output<PushPull>>,
    Spi<
        pac::SPI1,
        Spi1NoRemap,
        (
            gpioa::PA5<Alternate<PushPull>>,
            gpioa::PA6<Input<Floating>>,
            gpioa::PA7<Alternate<PushPull>>,
        ),
    >,
    gpioa::PA0<Output<PushPull>>,
    gpioa::PA4<Output<PushPull>>,
    gpioa::PA1<Output<PushPull>>,
    gpioa::PA9<Output<PushPull>>,
    qei::Qei<
        pac::TIM3,
        Tim3PartialRemap,
        (gpiob::PB4<Input<Floating>>, gpiob::PB5<Input<Floating>>),
    >,
> {
    // Take ownership over the raw flash and rcc devices and convert them into the corresponding
    // HAL structs
    let mut flash = dp.FLASH.constrain();

    let mut rcc = dp.RCC.constrain();

    // Freeze the configuration of all the clocks in the system and store the frozen frequencies in
    // `clocks`
    let clocks = rcc
        .cfgr
        .use_hse(8.mhz())
        .sysclk(72.mhz())
        .pclk1(36.mhz())
        .freeze(&mut flash.acr);

    // Acquire the GPIOC peripheral
    let mut gpioc = dp.GPIOC.split(&mut rcc.apb2);
    let mut gpioa = dp.GPIOA.split(&mut rcc.apb2);

    // Configure gpio C pin 13 as a push-pull output. The `crh` register is passed to the function
    // in order to configure the port. For pins 0-7, crl should be passed instead.
    let mut led = gpioc.pc13.into_push_pull_output(&mut gpioc.crh);

    let mut delay = Delay::new(cp.SYST, clocks);

    let gpiob = dp.GPIOB.split(&mut rcc.apb2);
    let mut afio = dp.AFIO.constrain(&mut rcc.apb2);

    let (_, _, pb4) = afio.mapr.disable_jtag(gpioa.pa15, gpiob.pb3, gpiob.pb4);

    let qei = Timer::tim3(dp.TIM3, &clocks, &mut rcc.apb1).qei(
        (pb4, gpiob.pb5),
        &mut afio.mapr,
        qei::QeiOptions::default(),
    );

    // SPI1
    let sck = gpioa.pa5.into_alternate_push_pull(&mut gpioa.crl);
    let miso = gpioa.pa6;
    let mosi = gpioa.pa7.into_alternate_push_pull(&mut gpioa.crl);

    let disp_cs = gpioa.pa0.into_push_pull_output(&mut gpioa.crl);
    let max_cs = gpioa.pa9.into_push_pull_output(&mut gpioa.crh);

    let rst = gpioa.pa1.into_push_pull_output(&mut gpioa.crl);
    let dc = gpioa.pa4.into_push_pull_output(&mut gpioa.crl);
    let mut disp_led = gpioa.pa8.into_push_pull_output(&mut gpioa.crh);

    disp_led.set_high().unwrap();

    let spi = Spi::spi1(
        dp.SPI1,
        (sck, miso, mosi),
        &mut afio.mapr,
        Mode {
            polarity: Polarity::IdleLow,
            phase: Phase::CaptureOnFirstTransition,
        },
        16.mhz(),
        clocks,
        &mut rcc.apb2,
    );

    App {
        delay: delay,
        board_led: led,
        spi: spi,
        disp_cs: disp_cs,
        disp_dc: dc,
        disp_rst: rst,
        max_cs: max_cs,
        qei: qei,
    }
}

impl<BLed, SPI, DispCS, DispDC, DispRST, MaxCS, QEI>
    App<BLed, SPI, DispCS, DispDC, DispRST, MaxCS, QEI>
where
    BLed: OutputPin,
    SPI: spi::Transfer<u8> + spi::Write<u8>,
    DispCS: OutputPin,
    DispDC: OutputPin,
    DispRST: OutputPin,
    MaxCS: OutputPin,
    QEI: Qei,
{
    pub fn run(mut self) -> ! {
        let mut disp =
            st7735_lcd::ST7735::new(self.spi, self.disp_dc, self.disp_rst, true, false, 160, 128);
        self.disp_cs.set_low().ok().unwrap();
        disp.init(&mut self.delay).unwrap();
        disp.set_orientation(&Orientation::LandscapeSwapped)
            .unwrap();

        let style_black = PrimitiveStyleBuilder::new()
            .fill_color(Rgb565::BLACK)
            .build();
        Rectangle::new(Point::new(0, 0), Point::new(160, 128))
            .into_styled(style_black)
            .draw(&mut disp)
            .unwrap();

        let bmp = Bmp::from_slice(include_bytes!("logo.bmp")).unwrap();
        let image = Image::new(&bmp, Point::new(16, 0));
        image.draw(&mut disp).unwrap();

        loop {}
    }
}
